# -*- coding: utf-8 -*-
"""

@author: natashav

This code takes an ice melt scenario and calculates the resulting elastic
fingerprint of sea level change. While it includes rotational effects, it does
not consider movement of shorelines in response to SL change.

"""

import numpy as np
from numpy.polynomial.legendre import leggauss
from scipy.io import loadmat
from scipy.interpolate import interp2d, LinearNDInterpolator
import matplotlib.pyplot as plt
import pyshtools as pysh
import time

#%% defining parameters, loading in topography, interpolating topo onto GL grid

# Parameters you will need
rhoi=917    # density of ice in kg/m^3
rhow=1000   # density of water in kg/m^3
a=6378100   # radius of the earth in m
Me=5.972e24 # mass of the earth in kg
g = 9.807   # acceleration due to gravity in m/s^2

# parameters for rotation
omega = 7.292e-5 # Earth's angular velocity in rad/s
kfa = 0.942 # fluid k tidal Love number at degree 2
k2te = 0.5989/2 # elastic component of k tidal Love number at degree 2
h2te = 0.605 # elastic component of h tidal Love number at degree 2
c_moment = 0.3307007*Me*a**2 # moment of inertia about rotation axis
a_moment = 0.3296145*Me*a**2 # moment about axes normal to rotation axis
# here we assume that the moments about both axes normal to the rotation
# axis are the same
b_moment = a_moment 
kfb = kfa

# Specify maximum degree to which spherical transformations should be done
maxdeg = 512

# Set up Gauss Legendre Grid
N = maxdeg
x, w = leggauss(N) # Gauss Quadrature points (x) and weights (w)
x_GL = np.arccos(x)*180/np.pi - 90 # cos(x_GL) are Quadrature points for P(cos(theta))
lon_GL = np.linspace(0,360,2*N+1)
lon_GL = lon_GL[:-1]
lon_out,lat_out = np.meshgrid(lon_GL,x_GL)

# Load in preloaded topography (includes ice on top of bedrock;
# from https://www.ngdc.noaa.gov/mgg/global/etopo2.html)
mat = loadmat("topo_Assignment.mat")
lat_topo = np.array(mat["lat_topo"])
lon_topo = np.array(mat["lon_topo"])
topo_ice = np.array(mat["topo_ice"])

# Interpolate topography onto Gauss Legendre Grid
x = interp2d(lon_topo, lat_topo, topo_ice)
topo_pres = x(lon_out[0,:], lat_out[:,0])

#%% a general function for any ice mask 
def SL_calc(topo_ice,lon_out,lat_out,maxdeg,w,rhoi,rhow,a,Me,ice_mask_path,fprint=False):
    '''
    Parameters
    ----------
    topo_ice : np array
        The ice topography array.
    lon_out : np array
        The longitude values for the output of the SL calculation.
        The ice topography array.
    lat_out : np array
        The latitude values for the output of the SL calculation.
    maxdeg : int
        The maximum degree to which the calculation is performed.
    w : np array
        The weights for Gauss quadrature.
    rhoi : int
        Ice density (kg/m^3).
    rhow: int
        Water density (kg/m^3).
    a : int
        Radius of the Earth (m).
    Me : int
        Mass of the Earth (kg).
    ice_path : string
        The path to the ice mask used in this calculation.
    fprint : boolean, optional
        Whether the function should return a normalized fingerprint instead
        of the calculated SL change. The default is False.

    Returns
    -------
    np array
        Depending on the value of fprint, either the fingerprint or
        SL change resulting from the given ice mask.
    '''
    
    # start timer for performance purposes
    start = time.perf_counter()
  
    # Set constants needed for pysh to perform the spherical harmonic 
    # transform 
    # these two numbers should be the same
    # equal to the degree at which you want to run the code
    lmax=maxdeg
    lreturn=maxdeg

    # Set-up grid and variables pysh uses for the spherical harmonic transform
    zero, w = pysh.expand.SHGLQ(lmax)                 # Transform variables
    latglq, longlq = pysh.expand.GLQGridCoord(lmax)   # Transform grid
    
    # create a GL grid for later use
    longlq_grid, latglq_grid = np.meshgrid(longlq, latglq)

    # Interpolate topo_ice from the input lon_out/lat_out grid onto the 
    # longlq/latglq grid
    # interpolate onto old grid
    f = interp2d(lon_out, lat_out, topo_ice)
    # project onto new grid, save as topo_pres
    topo_pres = f(longlq, latglq)
    
    C = topo_pres;

    C[C>0]=0
    C[C<0]=1

    # Do spherical harmonic transform of C into C_lm
    # we do this because we need the mask in spherical coordinates to calculate
    # the fingerprint
    c_lm = pysh.expand.SHExpandGLQC(C,w,zero,norm=1,csphase=1,lmax_calc=lreturn)
    
    # importing ice masks 
              
    ice_mask = np.loadtxt('./'+ice_mask_path)
    # the imported data is on a deg 512 GL grid
    lat_mask = np.loadtxt('./GLlat_512')
    lon_mask = np.loadtxt('./GLlon_512')
    
    # create a grid onto which to interpolate the ice mask data
    lon_grid, lat_grid = np.meshgrid(lon_mask, np.flip(lat_mask))
    
    # use scipy's LinearNDInterpolator bc too much data for interp2d
    # this does take a while to run! ~5 seconds I think
    # might need to change fill_value at some point
    z=LinearNDInterpolator(np.transpose(np.vstack((lon_grid.flatten(), lat_grid.flatten()))),\
                           ice_mask.flatten(), fill_value=0)
    ice_mask_interp = z(longlq_grid,latglq_grid)
    
    # set first guess for uniform SL change
    eustatic = 0.001
    delta_S_guess = eustatic
            
    # Do spherical harmonic transform of mask
    ice_mask_lm = pysh.expand.SHExpandGLQC(ice_mask_interp,w,zero,norm=1,csphase=1,lmax_calc=lreturn)

    # Re-calculate C for use later in the code
    topoglq2 = f(longlq,latglq)
    cilm_topo = pysh.expand.SHExpandGLQC(topoglq2,w,zero,norm=1,csphase=1,lmax_calc=lreturn)
    zero_down, w_down = pysh.expand.SHGLQ(lreturn)
    topo_down_grid = pysh.expand.MakeGridGLQC(cilm_topo,zero_down,lmax=lreturn,norm=1,csphase=1,lmax_calc=lreturn)
    topo_down = np.sign(topo_down_grid).real*np.abs(topo_down_grid)
    C=np.copy(topo_down)
    C[C>0]=0
    C[C<0]=1

    ## Use first guess to calculate delta_I, mask of the (uniform) ice height change

    # calculate the area of the ocean and area of ice first
    # the 1st value of a spherical harmonic xxx is the area
    area_ocean = c_lm[0][0,0].real
    area_ice = ice_mask_lm[0][0,0].real

    # Calculate ice height change
    # eustatic change in ocean water * rhow * ocean area = mass of water (kg)
    # mass of water / rhoi = ice volume (m^3)
    # [volume = area * height]
    # ice volume / area of GrIS ice = height of ice (m)
    height_ic = -delta_S_guess * rhow * area_ocean / rhoi / area_ice

    # Calculate delta_I (project onto ice mask)
    delta_I = ice_mask_interp * height_ic

    # Spherical harmonic transform
    delta_I_lm = pysh.expand.SHExpandGLQC(delta_I,w,zero,norm=1,csphase=1,lmax_calc=lreturn)

    ## Import the Elastic Love numbers
    # Love numbers are dimensionless numbers that represent the way a plantery 
    # body responds to changes in mass loads on its surface
    lovenumbers=np.loadtxt('./lovenumbers_python.txt')

    love_h=lovenumbers[:,1]
    love_k=lovenumbers[:,2]

    l=np.zeros([2,maxdeg+1,maxdeg+1])
    k_l=np.zeros([2,maxdeg+1,maxdeg+1])
    h_l=np.zeros([2,maxdeg+1,maxdeg+1])
    i,j,k=l.shape
    for ii in np.arange(0,i):
        for jj in np.arange(1,j):
            for kk in np.arange(0,k):
                l[ii,jj,kk]=jj
    for ii in np.arange(0,i):
        for jj in np.arange(1,j):
            for kk in np.arange(0,k):
                k_l[ii,jj,kk]=love_k[jj-1]
    for ii in np.arange(0,i):
        for jj in np.arange(1,j):
            for kk in np.arange(0,k):
                h_l[ii,jj,kk]=love_h[jj-1]

    # Calculate Tl and El for use later in the code
    E_l=1+k_l-h_l
    T_l=(4*np.pi*a**3)/(Me*(2*l+1))

    ## Calculate the initial delta_S, the first guess eustatic ocean height change.
    delta_S = - rhoi / rhow * delta_I_lm[0][0,0].real / area_ocean * C

    # spherical harmonic transform
    delta_S_lm = pysh.expand.SHExpandGLQC(delta_S,w,zero,norm=1,csphase=1,lmax_calc=lreturn)
    
    # NEW: initial calculation of rotation signal
    # need the k elastic load Love number of degree 2
    # choosing the index-1 element of the list bc there's no degree-one
    # elastic k Love number bc we are assuming mass conservation
    k2le = love_k[1]
    
    # for now, we only compute \Lambda_{21}, so we only need L_{21}
    # load component (from Matsuyama et al., 2006)
    temp = rhow*delta_S_lm + rhoi*delta_I_lm

    # calculate 13 and 23 components of moment of inertia (Matsuyama)
    il_13 = Me*a/g*np.sqrt(10/3)*(4*np.pi*a**3*g/(5*Me))*temp[0,2,1].real 
    il_23 = -Me*a/g*np.sqrt(10/3)*(4*np.pi*a**3*g/(5*Me))*temp[0,2,1].imag
    
    # compute m1 and m2 (perturbations to rotation axis along axes normal to
    # rotation axis)
    m1 = 1/(c_moment-a_moment)*il_13*(1+k2le)/(1-k2te/kfa)
    m2 = 1/(c_moment-b_moment)*il_23*(1+k2le)/(1-k2te/kfb)
    # compute \Lambda_21 to first order (Milne)
    lambda_21 = a**2*omega**2/np.sqrt(30)*complex(m1,-m2)
    lambda_lm = np.zeros_like(delta_S_lm)
    lambda_lm[0,2,1] = lambda_21
    # include 2,-1 component bc it has same magnitude as 2,1 (Milne)
    lambda_lm[1,2,1] = -np.conj(lambda_21)
    
    # note: the last index doesn't matter, as long as it's in the bounds
    # of the axis - all values on that row are the same, since that row
    # varies in m and E_l is solely l-dependent
    delta_S_lm += lambda_lm*E_l[0,2,1]/g
    
    ## Create a loop to iteratively calculate delta_SL, global sea-level variation

    # convergence criterion
    epsilon = 1e-4;
    # set a maximum iteration - don't want the loop to go forever & crash
    maxiter = 10;

    for i in np.arange(1,maxiter):
        # calculate delta SL
        temp = rhoi * delta_I_lm + rhow * delta_S_lm
        
        # convert from spherical back to spatial coordinates
        delta_SL = pysh.expand.MakeGridGLQC((temp*T_l*E_l),zero_down,lmax=lreturn,norm=1,csphase=1,lmax_calc=lreturn) #TlEl is the wrong size - I'll come back to that later, old code: TlEl*temp
        
        # calculate RO (project onto ocean mask)
        delta_RO = np.multiply(delta_SL, C)
        #convert to spherical harmonic coordinates
        delta_RO_lm = pysh.expand.SHExpandGLQC(delta_RO,w,zero,norm=1,csphase=1,lmax_calc=lreturn)
        
        # calculate delta phi / g (eustatic sea-level change)
        delta_phi_g = - rhoi / rhow * delta_I_lm[0][0,0].real / area_ocean - delta_RO_lm[0][0,0].real / area_ocean

        # calculate rotation effect
        # calculate 13 and 23 components of moment of inertia (Matsuyama)
        il_13 = Me*a/g*np.sqrt(10/3)*(4*np.pi*a**3*g/(5*Me))*temp[0,2,1].real 
        il_23 = -Me*a/g*np.sqrt(10/3)*(4*np.pi*a**3*g/(5*Me))*temp[0,2,1].imag
          
        # compute m1 and m2 (perturbations to rotation axis along axes normal to
        # it - from JXM handwritten notes)
        m1 = 1/(c_moment-a_moment)*il_13*(1+k2le)/(1-k2te/kfa)
        m2 = 1/(c_moment-b_moment)*il_23*(1+k2le)/(1-k2te/kfb)
        # compute \Lambda_21 to first order (Milne and Mitrovica, 1998)
        lambda_21 = a**2*omega**2/np.sqrt(30)*complex(m1,-m2)        
        lambda_lm[0,2,1] = lambda_21
        # include 2,-1 component bc it has same magnitude as 2,1 (Milne)
        lambda_lm[1,2,1] = -np.conj(lambda_21)
        
        # sum the two terms from the nonrotating earth
        delta_SL_no_rot = delta_SL + delta_phi_g
        
        # convert to SH space and add the effect of rotation
        delta_SL_rot_lm = pysh.expand.SHExpandGLQC(delta_SL_no_rot,w,zero,norm=1,csphase=1,lmax_calc=lreturn)\
            + lambda_lm*(1+k2te-h2te)/g
        # come back to spatial coords - this is our sea level estimate
        delta_SL_rot = pysh.expand.MakeGridGLQC(delta_SL_rot_lm,zero_down,lmax=lreturn,norm=1,csphase=1,lmax_calc=lreturn)
        
        # apply the ocean mask
        delta_S_new = delta_SL_rot * C
        # convert the masked SL to SH coords for the next iteration
        delta_S_lm_new = pysh.expand.SHExpandGLQC(delta_S_new,w,zero,norm=1,csphase=1,lmax_calc=lreturn)
        
        # check convergence
        chi = np.max(np.abs(delta_S_new - delta_S))
        if chi < epsilon:
            # compute elapsed time, report iterations required
            end = time.perf_counter()
            print(ice_mask_path+' converged after ', str(i), ' iterations. Chi was ',\
                  str(np.real(chi)),\
                  ', time to convergence was {:.3f} seconds.'.format(end-start))
            print("")
            break

        delta_S = np.copy(delta_S_new)
        delta_S_lm = np.copy(delta_S_lm_new)

    # Calculate the final ocean height change from the loop
    sea_level = delta_SL_rot
    
    if fprint:
        fingerprint = sea_level.real/eustatic
        return fingerprint

    return sea_level.real

#%% plotting the SL change from an Antarctic ice melt profile from an 
# ISMIP6 experiment

AIS_115 = SL_calc(topo_ice,lon_topo, lat_topo, maxdeg, w, rhoi, rhow, a, Me,'outGL_Diff_115',fprint=True)

# set up discrete colorbar
cmap = plt.get_cmap('coolwarm_r',15)
cmap.set_over('white') # color for values above scale range
cmap.set_under('gray') # color for values below scale range

# plot the fingerprint
fig = plt.figure(figsize=(14,10))
plt.pcolor(AIS_115, vmin=0.,vmax=1.5,cmap=cmap)
# feel free to play around with the max/min (they should be equal though)
# and feel free to change the colormap
cbar = plt.colorbar()
plt.title('run 115 fingerprint')
plt.contour(topo_pres,levels=[0],colors='black')
plt.tick_params(bottom=False, top=False, labelbottom=False, right=False,\
                left=False, labelleft=False)

plt.show()

